'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from collections import deque

from haigha.message import Message
from haigha.writer import Writer
from haigha.frames import MethodFrame, HeaderFrame, ContentFrame
from haigha.classes import ProtocolClass

class BasicClass(ProtocolClass):
  '''
  Implements the AMQP Basic class
  '''

  def __init__(self, *args, **kwargs):
    super(BasicClass, self).__init__(*args, **kwargs)
    self.dispatch_map = {
      11 : self._recv_qos_ok,
      21 : self._recv_consume_ok,
      31 : self._recv_cancel_ok,
      50 : self._recv_return,
      60 : self._recv_deliver,
      71 : self._recv_get_response,   # see impl
      72 : self._recv_get_response,   # see impl
      111 : self._recv_recover_ok,
    }

    self._consumer_tag_id = 0
    self._pending_consumers = deque()
    self._consumer_cb = {}
    self._get_cb = deque()
    self._recover_cb = deque()
    self._cancel_cb = deque()

  @property
  def name(self):
    return 'basic'

  def _cleanup(self):
    '''
    Cleanup all the local data.
    '''
    self._pending_consumers = None
    self._consumer_cb = None
    self._get_cb = None
    self._recover_cb = None
    self._cancel_cb = None
    super(BasicClass,self)._cleanup()

  def _generate_consumer_tag(self):
    '''
    Generate the next consumer tag.

    The consumer tag is local to a channel, so two clients can use the
    same consumer tags.
    '''
    self._consumer_tag_id += 1
    return "channel-%d-%d"%( self.channel_id, self._consumer_tag_id )

  def qos(self, prefetch_size=0, prefetch_count=0, is_global=False):
    '''
    Set QoS on this channel.
    '''
    args = Writer()
    args.write_long(prefetch_size).\
      write_short(prefetch_count).\
      write_bit(is_global)
    self.send_frame( MethodFrame(self.channel_id, 60, 10, args) )

    self.channel.add_synchronous_cb( self._recv_qos_ok )

  def _recv_qos_ok(self, _method_frame):
    # No arguments, nothing to do
    pass

  def consume(self, queue, consumer, consumer_tag='', no_local=False,
        no_ack=True, exclusive=False, nowait=True, ticket=None, cb=None):
    '''
    Start a queue consumer. If `cb` is supplied, will be called when
    broker confirms that consumer is registered.
    '''
    nowait = nowait and self.allow_nowait() and not cb

    if nowait and consumer_tag=='':
      consumer_tag = self._generate_consumer_tag()

    args = Writer()
    args.write_short(ticket or self.default_ticket).\
      write_shortstr(queue).\
      write_shortstr(consumer_tag).\
      write_bits( no_local, no_ack, exclusive, nowait ).\
      write_table({}) # unused according to spec
    self.send_frame( MethodFrame(self.channel_id, 60, 20, args) )

    if not nowait:
      self._pending_consumers.append( (consumer,cb) )
      self.channel.add_synchronous_cb( self._recv_consume_ok )
    else:
      self._consumer_cb[ consumer_tag ] = consumer

  def _recv_consume_ok(self, method_frame):
    consumer_tag = method_frame.args.read_shortstr()
    consumer, cb = self._pending_consumers.popleft()

    self._consumer_cb[ consumer_tag ] = consumer
    if cb: cb()

  def cancel(self, consumer_tag='', nowait=True, consumer=None, cb=None):
    '''
    Cancel a consumer. Can choose to delete based on a consumer tag or the
    function which is consuming.  If deleting by function, take care to only
    use a consumer once per channel.
    '''
    if consumer:
      for (tag,func) in self._consumer_cb.iteritems():
        if func==consumer:
          consumer_tag = tag
          break

    nowait = nowait and self.allow_nowait() and not cb

    args = Writer()
    args.write_shortstr(consumer_tag).\
      write_bit(nowait)
    self.send_frame( MethodFrame(self.channel_id, 60, 30, args) )

    if not nowait:
      self._cancel_cb.append( cb )
      self.channel.add_synchronous_cb( self._recv_cancel_ok )
    else:
      try:
        del self._consumer_cb[consumer_tag]
      except KeyError:
        self.logger.warning( 'no callback registered for consumer tag " %s "', consumer_tag )

  def _recv_cancel_ok(self, method_frame):
    consumer_tag = method_frame.args.read_shortstr()
    try:
      del self._consumer_cb[consumer_tag]
    except KeyError:
      self.logger.warning( 'no callback registered for consumer tag " %s "', consumer_tag )

    cb = self._cancel_cb.popleft()
    if cb: cb()

  def publish(self, msg, exchange, routing_key, mandatory=False, immediate=False, ticket=None):
    '''
    publish a message.
    '''
    args = Writer()
    args.write_short(ticket or self.default_ticket).\
      write_shortstr(exchange).\
      write_shortstr(routing_key).\
      write_bits(mandatory, immediate)

    self.send_frame( MethodFrame(self.channel_id, 60, 40, args) )
    self.send_frame( HeaderFrame(self.channel_id, 60, 0, len(msg), msg.properties) )

    frame_max = self.channel.connection.frame_max
    for frame in ContentFrame.create_frames(self.channel_id, msg.body, frame_max):
      self.send_frame( frame )

  def return_msg(self, reply_code, reply_text, exchange, routing_key):
    '''
    Return a failed message.  Not named "return" because python interpreter
    can't deal with that.
    '''
    args = Writer()
    args.write_short( reply_code ).\
      write_shortstr( reply_text ).\
      write_shortstr( exchange ).\
      write_shortstr( routing_key )

    self.send_frame( MethodFrame(self.channel_id, 60, 50, args) )

  def _recv_return(self, _method_frame):
    # This seems like the right place to callback that the operation has
    # completed.
    pass

  def _recv_deliver(self, method_frame):
    msg = self._read_msg( method_frame, 
      with_consumer_tag=True, with_message_count=False )

    func = self._consumer_cb.get(msg.delivery_info['consumer_tag'], None)
    if func: func(msg)

  def get(self, queue, consumer=None, no_ack=True, ticket=None):
    '''
    Ask to fetch a single message from a queue.  If a consumer is supplied,
    the consumer will be called with either a Message argument, or None if 
    there is no message in queue. If a synchronous transport, Message or
    None is returned.
    '''
    args = Writer()
    args.write_short(ticket or self.default_ticket).\
      write_shortstr(queue).\
      write_bit(no_ack)

    self._get_cb.append( consumer )
    self.send_frame( MethodFrame(self.channel_id, 60, 70, args) )
    return self.channel.add_synchronous_cb( self._recv_get_response )

  def _recv_get_response(self, method_frame):
    '''
    Handle either get_ok or get_empty.  This is a hack because the synchronous
    callback stack is expecting one method to satisfy the expectation.  To
    keep that loop as tight as possible, work within those constraints. Use
    of get is not recommended anyway.
    '''
    if method_frame.method_id==71:
      return self._recv_get_ok( method_frame )
    elif method_frame.method_id==72:
      return self._recv_get_empty( method_frame )

  def _recv_get_ok(self, method_frame):
    msg = self._read_msg( method_frame, 
      with_consumer_tag=False, with_message_count=True )
    cb = self._get_cb.popleft()
    if cb: cb( msg )
    return msg

  def _recv_get_empty(self, _method_frame):
    # On empty, call back with None as the argument so that user code knows
    # it's empty and can take next action
    cb = self._get_cb.popleft()
    if cb: cb( None )

  def ack(self, delivery_tag, multiple=False):
    '''
    Acknowledge delivery of a message.  If multiple=True, acknowledge up-to
    and including delivery_tag.
    '''
    args = Writer()
    args.write_longlong(delivery_tag).\
      write_bit(multiple)

    self.send_frame( MethodFrame(self.channel_id, 60, 80, args) )

  def reject(self, delivery_tag, requeue=False):
    '''
    Reject a message.
    '''
    args = Writer()
    args.write_longlong( delivery_tag ).\
      write_bit( requeue )

    self.send_frame( MethodFrame(self.channel_id, 60, 90, args) )

  def recover_async(self, requeue=False):
    '''
    Redeliver all unacknowledged messaages on this channel.

    This method is deprecated in favour of the synchronous recover/recover-ok
    '''
    args = Writer()
    args.write_bit( requeue )

    self.send_frame( MethodFrame(self.channel_id, 60, 100, args) )

  def recover(self, requeue=False, cb=None):
    '''
    Ask server to redeliver all unacknowledged messages.
    '''
    args = Writer()
    args.write_bit( requeue )

    # The XML spec is incorrect; this method is always synchronous
    #  http://lists.rabbitmq.com/pipermail/rabbitmq-discuss/2011-January/010738.html
    self._recover_cb.append( cb )
    self.send_frame( MethodFrame(self.channel_id, 60, 110, args) )
    self.channel.add_synchronous_cb( self._recv_recover_ok )

  def _recv_recover_ok(self, _method_frame):
    cb = self._recover_cb.popleft()
    if cb: cb()

  def _read_msg(self, method_frame, with_consumer_tag=False, with_message_count=False):
    '''
    Support method to read a Message from the current frame buffer. Will return
    a Message, or re-queue current frames and raise a FrameUnderflow. Takes an
    optional argument on whether to read the consumer tag so it can be used
    for both deliver and get-ok.
    '''
    # No need to assert that is instance of Header or Content frames because
    # failure to access as such will result in exception that channel will
    # pick up and handle accordingly.
    header_frame = self.channel.next_frame()
    if header_frame:
      size = header_frame.size
      body = bytearray()
      rbuf_frames = deque([header_frame, method_frame])

      while len(body) < size:
        content_frame = self.channel.next_frame()
        if content_frame:
          rbuf_frames.appendleft( content_frame )
          body.extend( content_frame.payload.buffer() )
        else:
          self.channel.requeue_frames( rbuf_frames )
          raise self.FrameUnderflow()
    else:
      self.channel.requeue_frames( [method_frame] )
      raise self.FrameUnderflow()

    if with_consumer_tag:
      consumer_tag = method_frame.args.read_shortstr()
    delivery_tag = method_frame.args.read_longlong()
    redelivered = method_frame.args.read_bit()
    exchange = method_frame.args.read_shortstr()
    routing_key = method_frame.args.read_shortstr()
    if with_message_count:
      message_count = method_frame.args.read_long()

    delivery_info = {
      'channel': self.channel,
      'delivery_tag': delivery_tag,
      'redelivered': redelivered,
      'exchange': exchange,
      'routing_key': routing_key,
    }
    if with_consumer_tag:
      delivery_info['consumer_tag'] = consumer_tag
    if with_message_count:
      delivery_info['message_count'] = message_count

    return Message( body=body, delivery_info=delivery_info, **header_frame.properties )



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.frames.frame import Frame
from haigha.writer import Writer

class HeartbeatFrame(Frame):
  '''
  Frame for heartbeats.
  '''

  @classmethod
  def type(cls):
    # NOTE: The PDF spec say this should be 4 but the xml spec say it should be 8
    #       RabbitMQ seems to implement this as 8, but maybe that's a difference
    #       between 0.8 and 0.9 protocols
    # Using Rabbit 2.1.1 and protocol 0.9.1 it seems that 8 is indeed the correct type @AW
    # PDF spec: http://www.amqp.org/confluence/download/attachments/720900/amqp0-9-1.pdf?version=1&modificationDate=1227526523000
    # XML spec: http://www.amqp.org/confluence/download/attachments/720900/amqp0-9-1.xml?version=1&modificationDate=1227526672000
    # This is addressed in http://dev.rabbitmq.com/wiki/Amqp091Errata#section_29
    return 8
  
  @classmethod
  def parse(self, channel_id, payload):
    return HeartbeatFrame( channel_id )

  def write_frame(self, buf):
    writer = Writer(buf)
    writer.write_octet( self.type() )
    writer.write_short( self.channel_id )
    writer.write_long( 0 )
    writer.write_octet( 0xce )

HeartbeatFrame.register()



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from protocol_class import ProtocolClass
from basic_class import BasicClass
from channel_class import ChannelClass
from exchange_class import ExchangeClass
from queue_class import QueueClass
from transaction_class import TransactionClass



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai
from datetime import datetime
from io import BytesIO
from decimal import Decimal

from haigha.reader import Reader
import struct
import operator

class ReaderTest(Chai):

  def test_init(self):
    ba = Reader(bytearray('foo'))
    assert_true( isinstance(ba._input, buffer) )
    assert_equals( 'foo', str(ba._input) )
    assert_equals( 0, ba._start_pos )
    assert_equals( 0, ba._pos )
    assert_equals( 3, ba._end_pos )

    s = Reader('foo')
    assert_true( isinstance(s._input, buffer) )
    assert_equals( 'foo', str(s._input) )

    u = Reader(u'D\xfcsseldorf')
    assert_true( isinstance(u._input, buffer) )
    assert_equals( 'D\xc3\xbcsseldorf', str(u._input) )

    b = BytesIO('foo')
    i = Reader(b)
    assert_true( isinstance(i._input, buffer) )
    assert_equals( 'foo', str(i._input) )

    src = Reader( 'foo' )
    r = Reader( src )
    assert_true( isinstance(r._input, buffer) )
    assert_equals( id(src._input), id(r._input) )
    assert_equals( 0, r._start_pos )
    assert_equals( 3, r._end_pos )

    src = Reader( 'hello world' )
    r = Reader( src, 3, 5 )
    assert_true( isinstance(r._input, buffer) )
    assert_equals( id(src._input), id(r._input) )
    assert_equals( 3, r._start_pos )
    assert_equals( 8, r._end_pos )
    assert_equals( 3, r._pos )

    assert_raises( ValueError, Reader, 1 )

  def test_str(self):
    assert_equals( '\\x66\\x6f\\x6f', str(Reader('foo')) )

  def test_tell(self):
    r = Reader('')
    r._pos = 'foo'
    assert_equals( 'foo', r.tell() )

  def test_seek_whence_zero(self):
    r = Reader('', 3)
    assert_equals( 3, r._pos )
    r.seek( 5 )
    assert_equals( 8, r._pos )

  def test_seek_whence_one(self):
    r = Reader('')
    r._pos = 2
    r.seek( 5, 1 )
    assert_equals( 7, r._pos )

  def test_seek_whence_two(self):
    r = Reader('foo bar')
    r.seek( -3, 2 )
    assert_equals( 3, r._pos )

  def test_check_underflow(self):
    r = Reader('')
    r._pos = 0
    r._end_pos = 5
    r._check_underflow( 3 )
    assert_raises( Reader.BufferUnderflow, r._check_underflow, 8 )

  def test_check_len(self):
    r = Reader('foo bar')
    self.assert_equals( 7, len(r) )
    r = Reader('foo bar', 3)
    self.assert_equals( 4, len(r) )

  def test_buffer(self):
    r = Reader('hello world', 3, 5)
    self.assert_equals( buffer('lo wo'), r.buffer() )

  def test_read(self):
    b = Reader('foo')
    assert_equals('foo', b.read(3) )
    
    b = Reader('foo')
    assert_equals('fo', b.read(2) )

    b = Reader('foo')
    assert_raises( Reader.BufferUnderflow, b.read, 4 )

  def test_read_bit(self):
    b = Reader('\x01')
    assert_true( b.read_bit() )
    
    b = Reader('\x00')
    assert_false( b.read_bit() )
    
    b = Reader('\x02')
    assert_false( b.read_bit() )

    b = Reader('')
    assert_raises( Reader.BufferUnderflow, b.read_bit )
  
  def test_read_bits(self):
    b = Reader('\x01')
    assert_equals( [True], b.read_bits(1) )
    
    b = Reader('\x00')
    assert_equals( [False], b.read_bits(1) )
    
    b = Reader('\x02')
    assert_equals( [False,True], b.read_bits(2) )
    
    b = Reader('\x02')
    assert_equals( [False,True,False,False,False,False,False,False], b.read_bits(8) )

    b = Reader('\x00')
    assert_raises( ValueError, b.read_bits, 9 )
    assert_raises( ValueError, b.read_bits, -1 )
    assert_equals( [], b.read_bits(0) )

    b = Reader('')
    assert_raises( Reader.BufferUnderflow, b.read_bits, 2 )

  def test_read_octet(self):
    b = Reader('\xff')
    assert_equals( 255, b.read_octet() )
    assert_raises( Reader.BufferUnderflow, b.read_octet )

  def test_read_short(self):
    b = Reader('\xff\x00')
    assert_equals( 65280, b.read_short() )
    assert_raises( Reader.BufferUnderflow, b.read_short )
  
  def test_read_long(self):
    b = Reader('\xff\x00\xff\x00')
    assert_equals( 4278255360, b.read_long() )
    assert_raises( Reader.BufferUnderflow, b.read_long )
  
  def test_read_longlong(self):
    b = Reader('\xff\x00\xff\x00\xff\x00\xff\x00')
    assert_equals( 18374966859414961920L, b.read_longlong() )
    assert_raises( Reader.BufferUnderflow, b.read_longlong )

  def test_read_shortstr(self):
    b = Reader('\x05hello')
    assert_equals( 'hello', b.read_shortstr() )
    assert_raises( Reader.BufferUnderflow, b.read_shortstr )
    
    b = Reader('\x0bD\xc3\xbcsseldorf')
    assert_equals( 'D\xc3\xbcsseldorf', b.read_shortstr() )
    
    b = Reader('\x05hell')
    assert_raises( Reader.BufferUnderflow, b.read_shortstr )

  def test_read_longstr(self):
    b = Reader('\x00\x00\x01\x00'+('a'*256))
    assert_equals( 'a'*256, b.read_longstr() )
    
    b = Reader('\x00\x00\x01\x00'+('a'*255))
    assert_raises( Reader.BufferUnderflow, b.read_longstr )

  def test_read_timestamp(self):
    b = Reader('\x00\x00\x00\x00\x4d\x34\xc4\x71')
    d = datetime(2011, 1, 17, 22, 36, 33)

    assert_equals( d, b.read_timestamp() )

  def test_read_table(self):
    # mock everything to keep this simple
    r = Reader('')
    expect( r.read_long ).returns( 42 )
    expect( r._check_underflow ).args( 42 )
    expect( r._field_shortstr ).returns( 'a' )
    expect( r._read_field ).returns( 3.14 ).side_effect( lambda: setattr(r, '_pos', 20) )
    expect( r._field_shortstr ).returns( 'b' )
    expect( r._read_field ).returns( 'pi' ).side_effect( lambda: setattr(r, '_pos', 42) )
    
    assert_equals( {'a':3.14,'b':'pi'}, r.read_table() )

  def test_read_field(self):
    r = Reader('Z')
    r.field_type_map['Z'] = mock()
    expect( r.field_type_map['Z'] ).args( r )
    
    r._read_field()

  def test_read_field_raises_fielderror_on_unknown_type(self):
    r = Reader('X')
    assert_raises( Reader.FieldError, r._read_field )

  def test_field_bool(self):
    r = Reader('\x00\x01\xf5')
    assert_false( r._field_bool() )
    assert_true( r._field_bool() )
    assert_true( r._field_bool() )
    assert_equals( 3, r._pos )

  def test_field_short_short_int(self):
    r = Reader( struct.pack('bb', 5, -5) )
    assert_equals( 5, r._field_short_short_int() )
    assert_equals( -5, r._field_short_short_int() )
    assert_equals( 2, r._pos )
  
  def test_field_short_short_uint(self):
    r = Reader( struct.pack('BB', 5, 255) )
    assert_equals( 5, r._field_short_short_uint() )
    assert_equals( 255, r._field_short_short_uint() )
    assert_equals( 2, r._pos )

  def test_field_short_int(self):
    r = Reader( struct.pack('>hh', 256, -256) )
    assert_equals( 256, r._field_short_int() )
    assert_equals( -256, r._field_short_int() )
    assert_equals( 4, r._pos )

  def test_field_short_uint(self):
    r = Reader( struct.pack('>HH', 256, 512) )
    assert_equals( 256, r._field_short_uint() )
    assert_equals( 512, r._field_short_uint() )
    assert_equals( 4, r._pos )

  def test_field_long_int(self):
    r = Reader( struct.pack('>ii', 2**16, -2**16) )
    assert_equals( 2**16, r._field_long_int() )
    assert_equals( -2**16, r._field_long_int() )
    assert_equals( 8, r._pos )

  def test_field_long_uint(self):
    r = Reader( struct.pack('>I', 2**32-1) )
    assert_equals( 2**32-1, r._field_long_uint() )
    assert_equals( 4, r._pos )

  def test_field_long_long_int(self):
    r = Reader( struct.pack('>qq', 2**32, -2**32) )
    assert_equals( 2**32, r._field_long_long_int() )
    assert_equals( -2**32, r._field_long_long_int() )
    assert_equals( 16, r._pos )

  def test_field_long_long_uint(self):
    r = Reader( struct.pack('>Q', 2**64-1) )
    assert_equals( 2**64-1, r._field_long_long_uint() )
    assert_equals( 8, r._pos )

  def test_field_float(self):
    r = Reader( struct.pack('>f', 3.1421) )
    assert_almost_equals( 3.1421, r._field_float(), 4 )
    assert_equals( 4, r._pos )

  def test_field_double(self):
    r = Reader( struct.pack('>d', 8675309.1138) )
    assert_almost_equals( 8675309.1138, r._field_double(), 4 )
    assert_equals( 8, r._pos )

  def test_field_decimal(self):
    r = Reader( struct.pack('>Bi', 2, 5) )
    assert_equals( Decimal('0.05'), r._field_decimal() )
    assert_equals( 5, r._pos )
    
    r = Reader( struct.pack('>Bi', 2, -5) )
    assert_equals( Decimal('-0.05'), r._field_decimal() )
    assert_equals( 5, r._pos )

  def test_field_shortstr(self):
    r = Reader( '\x05hello' )
    assert_equals( 'hello', r._field_shortstr() )
    assert_equals( 6, r._pos )

  def test_field_longstr(self):
    r = Reader( '\x00\x00\x01\x00'+('a'*256) )
    assert_equals( 'a'*256, r._field_longstr() )
    assert_equals( 260, r._pos )

  def test_field_array(self):
    # easier to mock the behavior here
    r = Reader('')
    expect( r.read_long ).returns( 42 )
    expect( r._read_field ).returns( 3.14 ).side_effect( lambda: setattr(r, '_pos', 20) )
    expect( r._read_field ).returns( 'pi' ).side_effect( lambda: setattr(r, '_pos', 42) )
    
    assert_equals( [3.14,'pi'], r._field_array() )
  
  def test_field_timestamp(self):
    b = Reader('\x00\x00\x00\x00\x4d\x34\xc4\x71')
    d = datetime(2011, 1, 17, 22, 36, 33)

    assert_equals( d, b._field_timestamp() )

  def test_field_bytearray(self):
    b = Reader('\x00\x00\x00\x03\x04\x05\x06')
    assert_equals( bytearray('\x04\x05\x06'), b._field_bytearray() )

  def test_field_none(self):
    b = Reader('')
    assert_equals( None, b._field_none() )
  
  def test_field_type_map_rabbit_errata(self):
    # http://dev.rabbitmq.com/wiki/Amqp091Errata#section_3
    assert_equals(
      {
        't' : Reader._field_bool.im_func,
        'b' : Reader._field_short_short_int.im_func,
        's' : Reader._field_short_int.im_func,
        'I' : Reader._field_long_int.im_func,
        'l' : Reader._field_long_long_int.im_func,
        'f' : Reader._field_float.im_func,
        'd' : Reader._field_double.im_func,
        'D' : Reader._field_decimal.im_func,
        'S' : Reader._field_longstr.im_func,
        'T' : Reader._field_timestamp.im_func,
        'F' : Reader.read_table.im_func,
        'V' : Reader._field_none.im_func,
        'x' : Reader._field_bytearray.im_func,
      }, Reader.field_type_map )

  #def test_field_type_map_091_spec(self):
  #  assert_equals(
  #    {
  #      't' : Reader._field_bool.im_func,
  #      'b' : Reader._field_short_short_int.im_func,
  #      'B' : Reader._field_short_short_uint.im_func,
  #      'U' : Reader._field_short_int.im_func,
  #      'u' : Reader._field_short_uint.im_func,
  #      'I' : Reader._field_long_int.im_func,
  #      'i' : Reader._field_long_uint.im_func,
  #      'L' : Reader._field_long_long_int.im_func,
  #      'l' : Reader._field_long_long_uint.im_func,
  #      'f' : Reader._field_float.im_func,
  #      'd' : Reader._field_double.im_func,
  #      'D' : Reader._field_decimal.im_func,
  #      's' : Reader._field_shortstr.im_func,
  #      'S' : Reader._field_longstr.im_func,
  #      'A' : Reader._field_array.im_func,
  #      'T' : Reader._field_timestamp.im_func,
  #      'F' : Reader.read_table.im_func,
  #    }, Reader.field_type_map )




'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai
from datetime import datetime
from decimal import Decimal

from haigha.writer import Writer


class WriterTest(Chai):

  def test_init(self):
    w = Writer()
    assert_equals( bytearray(), w._output_buffer )

    w = Writer(buf='foo')
    assert_equals( 'foo', w._output_buffer )

  def test_str(self):
    w = Writer(bytearray('\x03\xfb\x4d'))
    assert_equals( '\\x03\\xfb\\x4d', str(w) )
    assert_equals( '\\x03\\xfb\\x4d', repr(w) )

  def test_eq(self):
    assert_equals( Writer(bytearray('foo')), Writer(bytearray('foo')) )
    assert_not_equals( Writer(bytearray('foo')), Writer(bytearray('bar')) )

  def test_buffer(self):
    b = bytearray('pbt')
    assert_true( b is Writer(b).buffer() )

  def test_write(self):
    w = Writer()
    assert_true( w is w.write( 'foo' ) )
    assert_equals( bytearray('foo'), w._output_buffer )

  def test_write_bits(self):
    w = Writer()
    assert_true( w is w.write_bits(False,True) )
    assert_equals( bytearray('\x02'), w._output_buffer )
    w = Writer()
    assert_true( w is w.write_bits(False,False,True,True,True) )
    assert_equals( bytearray('\x1c'), w._output_buffer )

    assert_raises( ValueError, w.write_bits, *((True,)*9) )

  def test_write_bit(self):
    w = Writer()
    assert_true( w is w.write_bit(True) )
    assert_equals( bytearray('\x01'), w._output_buffer )

  def test_write_octet(self):
    w = Writer()
    assert_true( w is w.write_octet(0) )
    assert_true( w is w.write_octet(255) )
    assert_equals( bytearray('\x00\xff'), w._output_buffer )

    assert_raises( ValueError, w.write_octet, -1 )
    assert_raises( ValueError, w.write_octet, 2**8 )

  def test_write_short(self):
    w = Writer()
    assert_true( w is w.write_short(0) )
    assert_true( w is w.write_short(2**16-2) )
    assert_equals( bytearray('\x00\x00\xff\xfe'), w._output_buffer )

    assert_raises( ValueError, w.write_short, -1 )
    assert_raises( ValueError, w.write_short, 2**16 )

  def test_write_short_at(self):
    w = Writer( bytearray('\x00'*6) )
    assert_true( w is w.write_short_at(2**16-1,2) )
    assert_equals( bytearray('\x00\x00\xff\xff\x00\x00'), w._output_buffer )
    
    assert_raises( ValueError, w.write_short_at, -1, 2 )
    assert_raises( ValueError, w.write_short_at, 2**16, 3 )

  def test_write_long(self):
    w = Writer()
    assert_true( w is w.write_long(0) )
    assert_true( w is w.write_long(2**32-2) )
    assert_equals( bytearray('\x00\x00\x00\x00\xff\xff\xff\xfe'), w._output_buffer )
    
    assert_raises( ValueError, w.write_long, -1 )
    assert_raises( ValueError, w.write_long, 2**32 )

  def test_write_long_at(self):
    w = Writer( bytearray('\x00'*8) )
    assert_true( w is w.write_long_at(2**32-1,2) )
    assert_equals( bytearray('\x00\x00\xff\xff\xff\xff\x00\x00'), w._output_buffer )
    
    assert_raises( ValueError, w.write_long_at, -1, 2 )
    assert_raises( ValueError, w.write_long_at, 2**32, 3 )

  def test_write_long_long(self):
    w = Writer()
    assert_true( w is w.write_longlong(0) )
    assert_true( w is w.write_longlong(2**64-2) )
    assert_equals( 
      bytearray('\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff\xff\xff\xff\xff\xff\xfe'),
      w._output_buffer )
    
    assert_raises( ValueError, w.write_longlong, -1 )
    assert_raises( ValueError, w.write_longlong, 2**64 )

  def test_write_shortstr(self):
    w = Writer()
    assert_true( w is w.write_shortstr('') )
    assert_equals( bytearray('\x00'), w._output_buffer )
    w = Writer()
    assert_true( w is w.write_shortstr('a'*255) )
    assert_equals( bytearray('\xff'+('a'*255)), w._output_buffer )
    w = Writer()
    assert_true( w is w.write_shortstr( 'Au\xc3\x9ferdem'.decode('utf8') ) )
    assert_equals( bytearray('\x09Au\xc3\x9ferdem'), w._output_buffer )

    assert_raises( ValueError, w.write_shortstr, 'a'*256 )

  def test_write_longstr(self):
    # We can't actually build a string 2**32 long, so can't also test the
    # valueerror without mocking
    w = Writer()
    assert_true( w is w.write_longstr('') )
    assert_equals( bytearray('\x00\x00\x00\x00'), w._output_buffer )
    w = Writer()
    assert_true( w is w.write_longstr('a' * (2**16)) )
    assert_equals( bytearray('\x00\x01\x00\x00'+('a' * 2**16)), w._output_buffer )
    w = Writer()
    assert_true( w is w.write_longstr( 'Au\xc3\x9ferdem'.decode('utf8') ) )
    assert_equals( bytearray('\x00\x00\x00\x09Au\xc3\x9ferdem'), w._output_buffer )

    # TODO: mock valueerror when chai fixes the '__len__' problems with Mocks
    # since we can't actually create a long-enough string

  def test_write_timestamp(self):
    w = Writer()
    w.write_timestamp( datetime(2011, 1, 17, 22, 36, 33) )

    assert_equals( '\x00\x00\x00\x00\x4d\x34\xc4\x71', w._output_buffer )

  def test_write_table(self):
    w = Writer()
    expect( w._write_item ).args( 'a', 'foo' ).any_order().side_effect( 
      lambda *args: (setattr(w,'_pos',20), w._output_buffer.extend('afoo')) )
    expect( w._write_item ).args( 'b', 'bar' ).any_order().side_effect( 
      lambda *args: (setattr(w,'_pos',20), w._output_buffer.extend('bbar')) )

    assert_true( w is w.write_table({'a':'foo','b':'bar'}) )
    assert_equals( '\x00\x00\x00\x08', w._output_buffer[:4] )
    assert_equals( 12, len(w._output_buffer) )

  def test_write_item(self):
    w = Writer()
    expect( w.write_shortstr ).args( 'key' )
    expect( w._write_field ).args( 'value' )
    w._write_item( 'key', 'value' )

  def test_write_field(self):
    w = Writer()
    unknown = mock()
    expect( w._field_none ).args( unknown )
    w._write_field( unknown )
    
    Writer.field_type_map[type(unknown)] = unknown
    expect( unknown ).args( w, unknown )
    w._write_field( unknown )

  def test_field_bool(self):
    w = Writer()
    w._field_bool( True )
    w._field_bool( False )
    assert_equals( 't\x01t\x00', w._output_buffer )

  def test_field_int(self):
    w = Writer()
    w._field_int( -2**15 )
    w._field_int( 2**15-1 )
    assert_equals( 's\x80\x00s\x7f\xff', w._output_buffer )
    
    w = Writer()
    w._field_int( -2**31 )
    w._field_int( 2**31-1 )
    assert_equals( 'I\x80\x00\x00\x00I\x7f\xff\xff\xff', w._output_buffer )
    
    w = Writer()
    w._field_int( -2**63 )
    w._field_int( 2**63-1 )
    assert_equals( 
      'l\x80\x00\x00\x00\x00\x00\x00\x00l\x7f\xff\xff\xff\xff\xff\xff\xff', 
      w._output_buffer )

  def test_field_double(self):
    w = Writer()
    w._field_double( 3.1457923 )
    assert_equals( 'd\x40\x09\x2a\x95\x27\x44\x11\xa8', w._output_buffer )

  def test_field_decimal(self):
    w = Writer()
    w._field_decimal( Decimal('1.50') )
    assert_equals( 'D\x02\x00\x00\x00\x96', w._output_buffer )
    
    w = Writer()
    w._field_decimal( Decimal('-1.50') )
    assert_equals( 'D\x02\xff\xff\xff\x6a', w._output_buffer )

  def test_field_str(self):
    w = Writer()
    w._field_str( 'foo' )
    assert_equals( 'S\x00\x00\x00\x03foo', w._output_buffer )

  def test_field_unicode(self):
    w = Writer()
    w._field_unicode( 'Au\xc3\x9ferdem'.decode('utf8') )
    assert_equals( 'S\x00\x00\x00\x09Au\xc3\x9ferdem', w._output_buffer )

  def test_field_timestamp(self):
    w = Writer()
    w._field_timestamp( datetime(2011, 1, 17, 22, 36, 33) )

    assert_equals( 'T\x00\x00\x00\x00\x4d\x34\xc4\x71', w._output_buffer )

  def test_field_table(self):
    w = Writer()
    expect( w.write_table ).args( {'foo':'bar'} ).side_effect(
      lambda *args: w._output_buffer.extend('tdata') )
    w._field_table( {'foo':'bar'} )

    assert_equals( 'Ftdata', w._output_buffer )

  def test_field_none(self):
    w = Writer()
    w._field_none( None )
    w._field_none( 'zomg' )
    assert_equals( 'VV', w._output_buffer )
  
  def test_field_bytearray(self):
    w = Writer()
    w._field_bytearray( bytearray('foo') )
    assert_equals( 'x\x00\x00\x00\x03foo', w._output_buffer )

  def test_field_iterable(self):
    w = Writer()
    expect( w._write_field ).args('la').side_effect( 
      lambda field: w._output_buffer.append('a') )
    expect( w._write_field ).args('lb').side_effect(
      lambda field: w._output_buffer.append('b') )
    w._field_iterable( ['la','lb'] )
    
    expect( w._write_field ).args('ta').side_effect( 
      lambda field: w._output_buffer.append('a') )
    expect( w._write_field ).args('tb').side_effect(
      lambda field: w._output_buffer.append('b') )
    w._field_iterable( ('ta','tb') )
    
    expect( w._write_field ).args('sa').any_order().side_effect( 
      lambda field: w._output_buffer.append('s') )
    expect( w._write_field ).args('sb').any_order().side_effect(
      lambda field: w._output_buffer.append('s') )
    w._field_iterable( set(('sa','sb')) )

    assert_equals( 'AabAabAss', w._output_buffer )

  def test_field_type_map(self):
    assert_equals(
      {
        bool      : Writer._field_bool.im_func,
        int       : Writer._field_int.im_func,
        long      : Writer._field_int.im_func,
        float     : Writer._field_double.im_func,
        Decimal   : Writer._field_decimal.im_func,
        str       : Writer._field_str.im_func,
        unicode   : Writer._field_unicode.im_func,
        datetime  : Writer._field_timestamp.im_func,
        dict      : Writer._field_table.im_func,
        None      : Writer._field_none.im_func,
        bytearray : Writer._field_bytearray.im_func,
      }, Writer.field_type_map )




'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai
from collections import deque

from haigha import channel
from haigha.channel import Channel, SyncWrapper
from haigha.exceptions import ChannelError, ChannelClosed
from haigha.classes import *
from haigha.frames import *
from haigha.classes import *

class SyncWrapperTest(Chai):

  def test_init(self):
    s = SyncWrapper('cb')
    assert_equals('cb', s._cb)
    assert_true( s._read )
    assert_equals(None, s._result)

  def test_eq_when_other_is_same_cb(self):
    s = SyncWrapper('cb')
    assert_equals( 'cb', s )
    assert_not_equals( 'bb', s )

  def test_eq_when_other_has_same_cb(self):
    s = SyncWrapper('cb')
    other = SyncWrapper('cb')
    another = SyncWrapper('bb')

    assert_equals( s, other )
    assert_not_equals( s, another )

  def test_call(self):
    cb = mock()
    s = SyncWrapper( cb )

    expect( cb ).args('foo', 'bar', hello='mars')
    s('foo', 'bar', hello='mars')
    assert_false( s._read )
    

class ChannelTest(Chai):

  def test_init(self):
    c = Channel('connection', 'id', {
      20: ChannelClass,
      40: ExchangeClass,
      50: QueueClass,
      60: BasicClass,
      90: TransactionClass,
    })
    self.assertEquals( 'connection', c._connection )
    self.assertEquals( 'id', c._channel_id )
    self.assertTrue( isinstance(c.channel, ChannelClass) )
    self.assertTrue( isinstance(c.exchange, ExchangeClass) )
    self.assertTrue( isinstance(c.queue, QueueClass) )
    self.assertTrue( isinstance(c.basic, BasicClass) )
    self.assertTrue( isinstance(c.tx, TransactionClass) )
    self.assertEquals( c._class_map[20], c.channel )
    self.assertEquals( c._class_map[40], c.exchange )
    self.assertEquals( c._class_map[50], c.queue )
    self.assertEquals( c._class_map[60], c.basic )
    self.assertEquals( c._class_map[90], c.tx )
    self.assertEquals( deque([]), c._pending_events )
    self.assertEquals( deque([]), c._frame_buffer )
    assert_equals( set([]), c._open_listeners )
    assert_equals( set([]), c._close_listeners )
    assert_false( c._closed )
    assert_equals(
      {
        'reply_code'    : 0,
        'reply_text'    : 'first connect',
        'class_id'      : 0,
        'method_id'     : 0
      }, c._close_info )
    assert_true( c._active )

  def test_properties(self):
    connection = mock()
    connection.logger = 'logger'

    c = Channel(connection, 'id', {})
    c._closed = 'yes'
    c._close_info = 'ithappened'
    c._active = 'record'

    assertEquals( connection, c.connection )
    assertEquals( 'id', c.channel_id )
    assertEquals( 'logger', c.logger )
    assert_equals( 'yes', c.closed )
    assert_equals( 'ithappened', c.close_info )
    assert_equals( 'record', c.active )
    
    c._closed = False
    assert_equals( None, c.close_info )

  def test_add_open_listener(self):
    c = Channel(None,None,{})
    c.add_open_listener('foo')
    assert_equals( set(['foo']), c._open_listeners )

  def test_remove_open_listener(self):
    c = Channel(None,None,{})
    c.add_open_listener('foo')
    c.remove_open_listener('foo')
    c.remove_open_listener('bar')
    assert_equals( set([]), c._open_listeners )

  def test_notify_open_listeners(self):
    c = Channel(None,None,{})
    cb1 = mock()
    cb2 = mock()
    c._open_listeners = set([cb1,cb2])
    expect( cb1 ).args( c )
    expect( cb2 ).args( c )
    c._notify_open_listeners()

  def test_add_close_listener(self):
    c = Channel(None,None,{})
    c.add_close_listener('foo')
    assert_equals( set(['foo']), c._close_listeners )

  def test_remove_close_listener(self):
    c = Channel(None,None,{})
    c.add_close_listener('foo')
    c.remove_close_listener('foo')
    c.remove_close_listener('bar')
    assert_equals( set([]), c._close_listeners )

  def test_notify_close_listeners(self):
    c = Channel(None,None,{})
    cb1 = mock()
    cb2 = mock()
    c._close_listeners = set([cb1,cb2])
    expect( cb1 ).args( c )
    expect( cb2 ).args( c )
    c._notify_close_listeners()

  def test_open(self):
    c = Channel(None,None,{})
    expect( mock(c,'channel').open )
    c.open()
 
  def test_active(self):
    c = Channel(None, None,{})
    expect( mock(c,'channel').open )
    c.open()
    assertTrue(c.active)

  def test_close_with_no_args(self):
    c = Channel(None,None,{})
    expect( mock(c,'channel').close ).args( 0, '', 0, 0 )
    c.close()
  
  def test_close_with_args(self):
    c = Channel(None,None,{})
    expect( mock(c,'channel').close ).args(1, 'two', 3, 4)
    expect( c.channel.close ).args(1, 'two', 3, 4)

    c.close(1, 'two', 3, 4)
    c.close(reply_code=1, reply_text='two', class_id=3, method_id=4)

  def test_close_when_channel_attr_cleared(self):
    c = Channel(None,None,{})
    assert_false( hasattr(c, 'channel') )
    c.close()

  def test_publish(self):
    c = Channel(None,None,{})
    expect( mock(c,'basic').publish ).args( 'arg1', 'arg2', foo='bar' )
    c.publish( 'arg1', 'arg2', foo='bar' )

  def test_publish_synchronous(self):
    c = Channel(None,None,{})
    expect( mock(c,'tx').select )
    expect( mock(c,'basic').publish ).args( 'arg1', 'arg2', foo='bar' )
    expect( c.tx.commit ).args( cb='a_cb' )

    c.publish_synchronous( 'arg1', 'arg2', foo='bar', cb='a_cb' )

  def test_dispatch(self):
    c = Channel(None,None,{})
    frame = mock()
    frame.class_id = 32
    klass = mock()
    
    c._class_map[32] = klass
    expect( klass.dispatch ).args( frame )
    c.dispatch( frame )

    frame.class_id = 33
    assert_raises( Channel.InvalidClass, c.dispatch, frame )

  def test_buffer_frame(self):
    c = Channel(None,None,{})
    c.buffer_frame( 'f1' )
    c.buffer_frame( 'f2' )
    assertEquals( deque(['f1', 'f2']), c._frame_buffer )
  
  def test_process_frames_when_no_frames(self):
    # Not that this should ever happen, but to be sure
    c = Channel(None,None,{})
    stub( c.dispatch )
    c.process_frames()

  def test_process_frames_stops_when_buffer_is_empty(self):
    c = Channel(None,None,{})
    f0 = MethodFrame('ch_id', 'c_id', 'm_id')
    f1 = MethodFrame('ch_id', 'c_id', 'm_id')
    c._frame_buffer = deque([ f0, f1 ])

    expect( c.dispatch ).args( f0 )
    expect( c.dispatch ).args( f1 )

    c.process_frames()
    assert_equals( deque(), c._frame_buffer )

  def test_process_frames_stops_when_frameunderflow_raised(self):
    c = Channel(None,None,{})
    f0 = MethodFrame('ch_id', 'c_id', 'm_id')
    f1 = MethodFrame('ch_id', 'c_id', 'm_id')
    c._frame_buffer = deque([ f0, f1 ])

    expect( c.dispatch ).args( f0 ).raises( ProtocolClass.FrameUnderflow )

    c.process_frames()
    assertEquals( f1, c._frame_buffer[0] )
  
  def test_process_frames_logs_and_closes_when_dispatch_error_raised(self):
    c = Channel(None,None,{})
    c._connection = mock()
    c._connection.logger = mock()
    
    f0 = MethodFrame(20, 30, 40)
    f1 = MethodFrame('ch_id', 'c_id', 'm_id')
    c._frame_buffer = deque([ f0, f1 ])

    expect( c.dispatch ).args( f0 ).raises( RuntimeError("zomg it broked") )
    expect( c.close ).args( 500, 'Failed to dispatch %s'%(str(f0)) )

    assert_raises( RuntimeError, c.process_frames )
    assertEquals( f1, c._frame_buffer[0] )
  
  def test_process_frames_logs_and_closes_when_systemexit_raised(self):
    c = Channel(None,None,{})
    c._connection = mock()
    c._connection.logger = mock()
    
    f0 = MethodFrame(20, 30, 40)
    f1 = MethodFrame('ch_id', 'c_id', 'm_id')
    c._frame_buffer = deque([ f0, f1 ])

    expect( c.dispatch ).args( f0 ).raises( SystemExit() )
    stub( c.close )

    assert_raises( SystemExit, c.process_frames )
    assertEquals( f1, c._frame_buffer[0] )

  def test_next_frame_with_a_frame(self):
    c = Channel(None,None,{})
    ch_id, c_id, m_id = 0, 1, 2
    f0 = MethodFrame(ch_id, c_id, m_id)
    f1 = MethodFrame(ch_id, c_id, m_id)
    c._frame_buffer = deque([ f0, f1 ])
    assertEquals(c.next_frame(), f0)

  def test_next_frame_with_no_frames(self):
    c = Channel(None,None,{})
    c._frame_buffer = deque()
    assertEquals(c.next_frame(), None)

  def test_requeue_frames(self):
    c = Channel(None,None,{})
    ch_id, c_id, m_id = 0, 1, 2
    f = [MethodFrame(ch_id, c_id, m_id) for i in xrange(4)]
    c._frame_buffer = deque(f[:2])

    c.requeue_frames(f[2:])
    assertEquals(c._frame_buffer, deque([f[i] for i in [3, 2, 0, 1]]))
  
  def test_send_frame_when_not_closed_no_flow_control_no_pending_events(self):
    conn = mock()
    c = Channel(conn, 32, {})

    expect( conn.send_frame ).args('frame')
    
    c.send_frame( 'frame' )

  def test_send_frame_when_not_closed_no_flow_control_pending_event(self):
    conn = mock()
    c = Channel(conn, 32, {})
    c._pending_events.append( 'cb' )

    c.send_frame( 'frame' )
    self.assertEquals( deque(['cb','frame']), c._pending_events )

  def test_send_frame_when_not_closed_and_flow_control(self):
    conn = mock()
    c = Channel(conn, 32, {})
    c._active = False

    method = MethodFrame(1,2,3)
    heartbeat = HeartbeatFrame()
    header = HeaderFrame(1,2,3,4)
    content = ContentFrame(1,'foo')
    
    expect( conn.send_frame ).args( method )
    expect( conn.send_frame ).args( heartbeat )

    c.send_frame( method )
    c.send_frame( heartbeat )
    self.assertRaises( Channel.Inactive, c.send_frame, header )
    self.assertRaises( Channel.Inactive, c.send_frame, content )

  def test_send_frame_when_closed_for_a_reason(self):
    conn = mock()
    c = Channel(conn, 32, {})
    c._closed = True
    c._close_info = {'reply_code':42, 'reply_text':'bad'}

    assertRaises( ChannelClosed, c.send_frame, 'frame' )
  
  def test_send_frame_when_closed_for_no_reason(self):
    conn = mock()
    c = Channel(conn, 32, {})
    c._closed = True
    c._close_info = {'reply_code':42, 'reply_text':''}

    assertRaises( ChannelClosed, c.send_frame, 'frame' )

  def test_add_synchronous_cb_when_transport_asynchronous(self):
    conn = mock()
    conn.synchronous = False
    c = Channel(conn,None,{})

    assertEquals( deque([]), c._pending_events )
    c.add_synchronous_cb( 'foo' )
    assertEquals( deque(['foo']), c._pending_events )

  def test_add_synchronous_cb_when_transport_synchronous(self):
    conn = mock()
    conn.synchronous = True
    c = Channel(conn,None,{})

    wrapper = mock()
    wrapper._read = True
    wrapper._result = 'done'

    expect( channel.SyncWrapper ).args( 'foo' ).returns( wrapper )
    expect( conn.read_frames )
    expect( conn.read_frames ).side_effect(
      lambda: setattr(wrapper, '_read', False) )
    
    assertEquals( deque([]), c._pending_events )
    assert_equals( 'done', c.add_synchronous_cb('foo') )

    # This is technically cleared in runtime, but assert that it's not cleared
    # in this method
    assertEquals( deque([wrapper]), c._pending_events )

  def test_clear_synchronous_cb_when_no_pending(self):
    c = Channel(None,None,{})
    stub( c._flush_pending_events )

    assertEquals( deque([]), c._pending_events )
    assert_equals( 'foo', c.clear_synchronous_cb('foo') )

  def test_clear_synchronous_cb_when_pending_cb_matches(self):
    c = Channel(None,None,{})
    c._pending_events = deque(['foo'])

    expect( c._flush_pending_events )

    assert_equals( 'foo', c.clear_synchronous_cb('foo') )
    assertEquals( deque([]), c._pending_events )

  def test_clear_synchronous_cb_when_pending_cb_doesnt_match_but_isnt_in_list(self):
    c = Channel(None,None,{})
    c._pending_events = deque(['foo'])

    expect( c._flush_pending_events )

    assert_equals( 'bar', c.clear_synchronous_cb('bar') )
    assertEquals( deque(['foo']), c._pending_events )

  def test_clear_synchronous_cb_when_pending_cb_doesnt_match_but_isnt_in_list(self):
    c = Channel(None,None,{})
    stub( c._flush_pending_events )
    c._pending_events = deque(['foo', 'bar'])

    assertRaises( ChannelError, c.clear_synchronous_cb, 'bar' )
    assertEquals( deque(['foo','bar']), c._pending_events )

  def test_flush_pending_events_flushes_all_leading_frames(self):
    conn = mock()
    c = Channel(conn,42,{})
    f1 = MethodFrame(1,2,3)
    f2 = MethodFrame(1,2,3)
    f3 = MethodFrame(1,2,3)
    c._pending_events = deque([f1, f2, 'cb', f3])

    expect( conn.send_frame ).args( f1 )
    expect( conn.send_frame ).args( f2 )

    c._flush_pending_events()
    assertEquals( deque(['cb',f3]), c._pending_events )

  def test_closed_cb_without_final_frame(self):
    c = Channel('connection',None,{
      20: ChannelClass,
      40: ExchangeClass,
      50: QueueClass,
      60: BasicClass,
      90: TransactionClass,
    })
    c._pending_events = 'foo'
    c._frame_buffer = 'foo'
    
    for val in c._class_map.values():
      expect( val._cleanup )
    expect( c._notify_close_listeners )

    c._closed_cb()
    assert_equals( deque([]), c._pending_events )
    assert_equals( deque([]), c._frame_buffer )
    assert_equals( None, c._connection )
    assert_false( hasattr(c,'channel') )
    assert_false( hasattr(c,'exchange') )
    assert_false( hasattr(c,'queue') )
    assert_false( hasattr(c,'basic') )
    assert_false( hasattr(c,'tx') )
    assert_equals( None, c._class_map )
    assert_equals( set(), c._close_listeners )

  def test_closed_cb_with_final_frame(self):
    conn = mock()
    c = Channel(conn,None,{})

    expect( conn.send_frame ).args('final')    
    for val in c._class_map.values():
      expect( val._cleanup )

    c._closed_cb('final')



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.classes import ProtocolClass
from haigha.frames import MethodFrame
from haigha.writer import Writer

class ChannelClass(ProtocolClass):
  '''
  Implements the AMQP Channel class
  '''
  
  def __init__(self, *args, **kwargs):
    super(ChannelClass, self).__init__(*args, **kwargs)
    self.dispatch_map = {
      11 : self._recv_open_ok,
      20 : self._recv_flow,
      21 : self._recv_flow_ok,
      40 : self._recv_close,
      41 : self._recv_close_ok,
    }
    self._flow_control_cb = None

  @property
  def name(self):
    return 'channel'

  def set_flow_cb(self, cb):
    '''
    Set a callback that will be called when the state of flow control has changed.
    The caller should use closures if they need to receive a handle to the channel
    on which flow control changes.
    '''
    self._flow_control_cb = cb
  
  def open(self):
    '''
    Open the channel for communication.
    '''
    args = Writer()
    args.write_shortstr('')
    self.send_frame( MethodFrame(self.channel_id, 20, 10, args) )
    self.channel.add_synchronous_cb( self._recv_open_ok )

  def _recv_open_ok(self, method_frame):
    '''
    Channel is opened.
    '''
    self.channel._notify_open_listeners()

  def activate(self):
    '''
    Activate this channel (disable flow control).
    '''
    if not self.channel.active:
      self._send_flow( True )

  def deactivate(self):
    '''
    Deactivate this channel (enable flow control).
    '''
    if self.channel.active:
      self._send_flow( False )

  def _send_flow(self, active):
    '''
    Send a flow control command.
    '''
    args = Writer()
    args.write_bit( active )
    self.send_frame( MethodFrame(self.channel_id, 20, 20, args) )
    self.channel.add_synchronous_cb( self._recv_flow_ok )

  def _recv_flow(self, method_frame):
    '''
    Receive a flow control command from the broker
    '''
    self.channel._active = method_frame.args.read_bit()
    
    args = Writer()
    args.write_bit( self.channel.active )
    self.send_frame( MethodFrame(self.channel_id, 20, 21, args) )

    if self._flow_control_cb is not None:
      self._flow_control_cb()

  def _recv_flow_ok(self, method_frame):
    '''
    Receive a flow control ack from the broker.
    '''
    self.channel._active = method_frame.args.read_bit()
    if self._flow_control_cb is not None:
      self._flow_control_cb()

  def close(self, reply_code=0, reply_text='', class_id=0, method_id=0):
    '''
    Close this channel.  Caller has the option of specifying the reason for
    closure and the class and method ids of the current frame in which an error
    occurred.  If in the event of an exception, the channel will be marked
    as immediately closed.  If channel is already closed, call is ignored.
    '''
    if not getattr(self, 'channel', None) or self.channel._closed: return

    self.channel._close_info = {
      'reply_code'    : reply_code,
      'reply_text'    : reply_text,
      'class_id'      : class_id,
      'method_id'     : method_id
    }

    # exceptions here likely due to race condition as connection is closing
    # cap the reply_text we send because it may be arbitrarily long
    try:
      args = Writer()
      args.write_short( reply_code )
      args.write_shortstr( reply_text[:255] )
      args.write_short( class_id )
      args.write_short( method_id )
      self.send_frame( MethodFrame(self.channel_id, 20, 40, args) )
      
      self.channel.add_synchronous_cb( self._recv_close_ok )
    finally:
      # Immediately set the closed flag so that no more frames can be sent
      # NOTE: in synchronous mode, by the time this is called we will have
      # already run self.channel._closed_cb and so the channel reference is
      # gone.
      if self.channel:
        self.channel._closed = True

  def _recv_close(self, method_frame):
    '''
    Receive a close command from the broker.
    '''
    self.channel._close_info = {
      'reply_code'    : method_frame.args.read_short(),
      'reply_text'    : method_frame.args.read_shortstr(),
      'class_id'      : method_frame.args.read_short(),
      'method_id'     : method_frame.args.read_short()
    }

    self.channel._closed = True
    self.channel._closed_cb( final_frame=MethodFrame(self.channel_id, 20, 41) )

  def _recv_close_ok(self, method_frame):
    '''
    Receive a close ack from the broker.
    '''
    self.channel._closed = True
    self.channel._closed_cb()



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai

from haigha.frames import heartbeat_frame, HeartbeatFrame, Frame

class HeartbeatFrameTest(Chai):

  def test_type(self):
    assert_equals(8, HeartbeatFrame.type() )

  def test_parse(self):
    frame = HeartbeatFrame.parse(42, 'payload')
    assert_true( isinstance(frame, HeartbeatFrame) )
    assert_equals( 42, frame.channel_id )

  def test_write_frame(self):
    w = mock()
    expect( mock(heartbeat_frame, 'Writer') ).args('buffer').returns( w )
    expect( w.write_octet ).args( 8 ).returns( w )
    expect( w.write_short ).args( 42 ).returns( w )
    expect( w.write_long ).args( 0 ).returns( w )
    expect( w.write_octet ).args( 0xce )

    frame = HeartbeatFrame(42)
    frame.write_frame( 'buffer' )



'''
Copyright (c) 2012, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai

from haigha.classes import protocol_class, ProtocolClass

class ProtocolClassTest(Chai):

  def test_dispatch_when_in_dispatch_map(self):
    ch = mock()
    frame = mock()
    frame.method_id = 42
    
    klass = ProtocolClass( ch )
    klass.dispatch_map = { 42: 'method' }

    with expect( ch.clear_synchronous_cb ).args('method').returns(mock()) as cb:
      expect( cb ).args( frame )

    klass.dispatch(frame)



'''
Copyright (c) 2012, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from collections import deque
import logging
from chai import Chai

from haigha.connections import rabbit_connection
from haigha.connections.rabbit_connection import *
from haigha.connection import Connection
from haigha.writer import Writer
from haigha.frames import *
from haigha.classes import *

class RabbitConnectionTest(Chai):

  def test_init(self):
    with expect( mock(rabbit_connection, 'super') ).args(is_arg(RabbitConnection), RabbitConnection).returns(mock()) as c:
      expect( c.__init__ ).args(class_map=var('classes'), foo='bar')

    rc = RabbitConnection(foo='bar')
    assert_equals( var('classes').value, {
      40 : RabbitExchangeClass,
      60 : RabbitBasicClass,
      85 : RabbitConfirmClass
    } )

class RabbitExchangeClassTest(Chai):

  def setUp(self):
    super(RabbitExchangeClassTest,self).setUp()
    ch = mock()
    ch.channel_id = 42
    ch.logger = mock()
    self.klass = RabbitExchangeClass( ch )

  def test_init(self):
    assert_equals( self.klass.dispatch_map[31], self.klass._recv_bind_ok )
    assert_equals( self.klass.dispatch_map[51], self.klass._recv_unbind_ok )
    assert_equals( deque(), self.klass._bind_cb )
    assert_equals( deque(), self.klass._unbind_cb )

  def test_declare_default_args(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(rabbit_connection, 'Writer') ).returns( w )
    expect( w.write_short ).args( self.klass.default_ticket ).returns(w)
    expect( w.write_shortstr ).args( 'exchange' ).returns( w )
    expect( w.write_shortstr ).args( 'topic' ).returns( w )
    expect( w.write_bits ).args( False, False, True, False, True ).returns( w )
    expect( w.write_table ).args( {} )
    expect( mock(rabbit_connection, 'MethodFrame') ).args(42, 40, 10, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    stub( self.klass.channel.add_synchronous_cb )

    self.klass.declare('exchange', 'topic')
    assert_equals( deque(), self.klass._declare_cb )

  def test_declare_with_args(self):
    w = mock()
    stub( self.klass.allow_nowait )
    expect( mock(rabbit_connection, 'Writer') ).returns( w )
    expect( w.write_short ).args( 't' ).returns(w)
    expect( w.write_shortstr ).args( 'exchange' ).returns( w )
    expect( w.write_shortstr ).args( 'topic' ).returns( w )
    expect( w.write_bits ).args( 'p', 'd', 'ad', 'yes', False ).returns( w )
    expect( w.write_table ).args( 'table' )
    expect( mock(rabbit_connection, 'MethodFrame') ).args(42, 40, 10, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_declare_ok )

    self.klass.declare('exchange', 'topic', passive='p', durable='d', 
      nowait=False, arguments='table', ticket='t', 
      auto_delete='ad', internal='yes')
    assert_equals( deque([None]), self.klass._declare_cb )

  def test_declare_with_cb(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(rabbit_connection, 'Writer') ).returns( w )
    expect( w.write_short ).args( 't' ).returns(w)
    expect( w.write_shortstr ).args( 'exchange' ).returns( w )
    expect( w.write_shortstr ).args( 'topic' ).returns( w )
    expect( w.write_bits ).args( 'p', 'd', True, False, False ).returns( w )
    expect( w.write_table ).args( 'table' )
    expect( mock(rabbit_connection, 'MethodFrame') ).args(42, 40, 10, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_declare_ok )

    self.klass.declare('exchange', 'topic', passive='p', durable='d', 
      nowait=True, arguments='table', ticket='t', cb='foo')
    assert_equals( deque(['foo']), self.klass._declare_cb )

  def test_bind_default_args(self):
    w = mock()
    
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(rabbit_connection, 'Writer') ).returns( w )
    expect( w.write_short ).args( 0 ).returns( w )
    expect( w.write_shortstr ).args( 'destination' ).returns( w )
    expect( w.write_shortstr ).args( 'source' ).returns( w )
    expect( w.write_shortstr ).args( '' ).returns( w )
    expect( w.write_bit ).args( True ).returns( w )
    expect( w.write_table ).args( {} )
    expect( mock(rabbit_connection, 'MethodFrame') ).args(42, 40, 30, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    self.klass.bind('destination', 'source')
    assert_equals( deque(), self.klass._bind_cb )

  def test_bind_with_args(self):
    w = mock()
    
    stub( self.klass.allow_nowait )
    expect( mock(rabbit_connection, 'Writer') ).returns( w )
    expect( w.write_short ).args( 't' ).returns(w)
    expect( w.write_shortstr ).args( 'destination' ).returns( w )
    expect( w.write_shortstr ).args( 'source' ).returns( w )
    expect( w.write_shortstr ).args( 'route' ).returns( w )
    expect( w.write_bit ).args( False ).returns( w )
    expect( w.write_table ).args( 'table' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_bind_ok )
    expect( mock(rabbit_connection, 'MethodFrame') ).args(42, 40, 30, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    self.klass.bind('destination', 'source', routing_key='route', 
      ticket='t', nowait=False, arguments='table')
    assert_equals( deque([None]), self.klass._bind_cb )

  def test_bind_with_cb(self):
    w = mock()
    
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(rabbit_connection, 'Writer') ).returns( w )
    expect( w.write_short ).args( 't' ).returns(w)
    expect( w.write_shortstr ).args( 'destination' ).returns( w )
    expect( w.write_shortstr ).args( 'source' ).returns( w )
    expect( w.write_shortstr ).args( 'route' ).returns( w )
    expect( w.write_bit ).args( False ).returns( w )
    expect( w.write_table ).args( 'table' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_bind_ok )
    expect( mock(rabbit_connection, 'MethodFrame') ).args(42, 40, 30, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    self.klass.bind('destination', 'source', routing_key='route', 
      ticket='t', arguments='table', cb='foo')
    assert_equals( deque(['foo']), self.klass._bind_cb )
  
  def test_recv_bind_ok_no_cb(self):
    self.klass._bind_cb = deque([None])
    self.klass._recv_bind_ok('frame')
    assert_equals( deque(), self.klass._bind_cb )
  
  def test_recv_bind_ok_with_cb(self):
    cb = mock()
    self.klass._bind_cb = deque([cb])
    expect( cb )
    self.klass._recv_bind_ok('frame')
    assert_equals( deque(), self.klass._bind_cb )

  def test_unbind_default_args(self):
    w = mock()
    
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(rabbit_connection, 'Writer') ).returns( w )
    expect( w.write_short ).args( 0 ).returns( w )
    expect( w.write_shortstr ).args( 'destination' ).returns( w )
    expect( w.write_shortstr ).args( 'source' ).returns( w )
    expect( w.write_shortstr ).args( '' ).returns( w )
    expect( w.write_bit ).args( True ).returns( w )
    expect( w.write_table ).args( {} )
    expect( mock(rabbit_connection, 'MethodFrame') ).args(42, 40, 40, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    self.klass.unbind('destination', 'source')
    assert_equals( deque(), self.klass._unbind_cb )

  def test_unbind_with_args(self):
    w = mock()
    
    stub( self.klass.allow_nowait )
    expect( mock(rabbit_connection, 'Writer') ).returns( w )
    expect( w.write_short ).args( 't' ).returns(w)
    expect( w.write_shortstr ).args( 'destination' ).returns( w )
    expect( w.write_shortstr ).args( 'source' ).returns( w )
    expect( w.write_shortstr ).args( 'route' ).returns( w )
    expect( w.write_bit ).args( False ).returns( w )
    expect( w.write_table ).args( 'table' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_unbind_ok )
    expect( mock(rabbit_connection, 'MethodFrame') ).args(42, 40, 40, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    self.klass.unbind('destination', 'source', routing_key='route', 
      ticket='t', nowait=False, arguments='table')
    assert_equals( deque([None]), self.klass._unbind_cb )

  def test_unbind_with_cb(self):
    w = mock()
    
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(rabbit_connection, 'Writer') ).returns( w )
    expect( w.write_short ).args( 't' ).returns(w)
    expect( w.write_shortstr ).args( 'destination' ).returns( w )
    expect( w.write_shortstr ).args( 'source' ).returns( w )
    expect( w.write_shortstr ).args( 'route' ).returns( w )
    expect( w.write_bit ).args( False ).returns( w )
    expect( w.write_table ).args( 'table' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_unbind_ok )
    expect( mock(rabbit_connection, 'MethodFrame') ).args(42, 40, 40, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    self.klass.unbind('destination', 'source', routing_key='route', 
      ticket='t', arguments='table', cb='foo')
    assert_equals( deque(['foo']), self.klass._unbind_cb )
  
  def test_recv_unbind_ok_no_cb(self):
    self.klass._unbind_cb = deque([None])
    self.klass._recv_unbind_ok('frame')
    assert_equals( deque(), self.klass._unbind_cb )
  
  def test_recv_unbind_ok_with_cb(self):
    cb = mock()
    self.klass._unbind_cb = deque([cb])
    expect( cb )
    self.klass._recv_unbind_ok('frame')
    assert_equals( deque(), self.klass._unbind_cb )

class RabbitBasicClassTest(Chai):

  def setUp(self):
    super(RabbitBasicClassTest,self).setUp()
    ch = mock()
    ch.channel_id = 42
    ch.logger = mock()
    self.klass = RabbitBasicClass( ch )

  def test_init(self):
    assert_equals( self.klass.dispatch_map[80], self.klass._recv_ack )
    assert_equals( self.klass.dispatch_map[120], self.klass._recv_nack )
    assert_equals( None, self.klass._ack_listener )
    assert_equals( None, self.klass._nack_listener )
    assert_equals( 0, self.klass._msg_id )
    assert_equals( 0, self.klass._last_ack_id )

  def test_set_ack_listener(self):
    self.klass.set_ack_listener('foo')
    assert_equals( 'foo', self.klass._ack_listener )

  def test_set_nack_listener(self):
    self.klass.set_nack_listener('foo')
    assert_equals( 'foo', self.klass._nack_listener )

  def test_publish_when_not_confirming(self):
    self.klass.channel.confirm._enabled = False
    with expect( mock(rabbit_connection, 'super') ).args(
    is_arg(RabbitBasicClass), RabbitBasicClass).returns(mock()) as klass:
      expect( klass.publish ).args( 'a', 'b', c='d' )

    assert_equals( 0, self.klass.publish('a','b', c='d') )
    assert_equals( 0, self.klass._msg_id )

  def test_publish_when_confirming(self):
    self.klass.channel.confirm._enabled = True
    with expect( mock(rabbit_connection, 'super') ).args(
    is_arg(RabbitBasicClass), RabbitBasicClass).returns(mock()) as klass:
      expect( klass.publish ).args( 'a', 'b', c='d' )

    assert_equals( 1, self.klass.publish('a','b', c='d') )
    assert_equals( 1, self.klass._msg_id )

  def test_recv_ack_no_listener(self):
    self.klass._recv_ack('frame')

  def test_recv_ack_with_listener_single_msg(self):
    self.klass._ack_listener = mock()
    frame = mock()
    expect( frame.args.read_longlong ).returns( 42 )
    expect( frame.args.read_bit ).returns( False )
    expect( self.klass._ack_listener ).args( 42 )
    
    self.klass._recv_ack( frame )
    assert_equals( 42, self.klass._last_ack_id )

  def test_recv_ack_with_listener_multiple_msg(self):
    self.klass._ack_listener = mock()
    self.klass._last_ack_id = 40
    frame = mock()
    expect( frame.args.read_longlong ).returns( 42 )
    expect( frame.args.read_bit ).returns( True )
    expect( self.klass._ack_listener ).args( 41 )
    expect( self.klass._ack_listener ).args( 42 )
    
    self.klass._recv_ack( frame )
    assert_equals( 42, self.klass._last_ack_id )
  
  def test_nack_default_args(self):
    w = mock()
    expect( mock( rabbit_connection, 'Writer' ) ).returns( w )
    expect( w.write_longlong ).args( 8675309 ).returns( w )
    expect( w.write_bits ).args( False, False )
    expect( mock( rabbit_connection, 'MethodFrame' ) ).args( 42, 60, 120, w ).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    self.klass.nack( 8675309 )
  
  def test_nack_with_args(self):
    w = mock()
    expect( mock( rabbit_connection, 'Writer' ) ).returns( w )
    expect( w.write_longlong ).args( 8675309 ).returns( w )
    expect( w.write_bits ).args( 'many', 'sure' )
    expect( mock( rabbit_connection, 'MethodFrame' ) ).args( 42, 60, 120, w ).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    self.klass.nack( 8675309, multiple='many', requeue='sure' )

  def test_recv_nack_no_listener(self):
    self.klass._recv_nack('frame')

  def test_recv_nack_with_listener_single_msg(self):
    self.klass._nack_listener = mock()
    frame = mock()
    expect( frame.args.read_longlong ).returns( 42 )
    expect( frame.args.read_bits ).args(2).returns( (False,False) )
    expect( self.klass._nack_listener ).args( 42, False )
    
    self.klass._recv_nack( frame )
    assert_equals( 42, self.klass._last_ack_id )

  def test_recv_nack_with_listener_multiple_msg(self):
    self.klass._nack_listener = mock()
    self.klass._last_ack_id = 40
    frame = mock()
    expect( frame.args.read_longlong ).returns( 42 )
    expect( frame.args.read_bits ).args(2).returns( (True,True) )
    expect( self.klass._nack_listener ).args( 41, True )
    expect( self.klass._nack_listener ).args( 42, True )
    
    self.klass._recv_nack( frame )
    assert_equals( 42, self.klass._last_ack_id )

class RabbitConfirmClassTest(Chai):

  def setUp(self):
    super(RabbitConfirmClassTest,self).setUp()
    ch = mock()
    ch.channel_id = 42
    ch.logger = mock()
    self.klass = RabbitConfirmClass( ch )

  def test_init(self):
    assert_equals( {11:self.klass._recv_select_ok}, self.klass.dispatch_map )
    assert_false( self.klass._enabled )
    assert_equals( deque(), self.klass._select_cb )

  def test_name(self):
    assert_equals( 'confirm', self.klass.name )

  def test_select_when_not_enabled_and_no_cb(self):
    self.klass._enabled = False
    w = mock()
    expect( mock( rabbit_connection, 'Writer' ) ).returns( w )
    expect( self.klass.allow_nowait ).returns( True )
    expect( w.write_bit ).args( True )
    expect( mock(rabbit_connection, 'MethodFrame') ).args(42, 85, 10, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    stub( self.klass.channel.add_synchronous_cb )

    self.klass.select()
    assert_true( self.klass._enabled )
    assert_equals( deque(), self.klass._select_cb )

  def test_select_when_not_enabled_and_no_cb_but_synchronous(self):
    self.klass._enabled = False
    w = mock()
    expect( mock( rabbit_connection, 'Writer' ) ).returns( w )
    stub( self.klass.allow_nowait )
    expect( w.write_bit ).args( False )
    expect( mock(rabbit_connection, 'MethodFrame') ).args(42, 85, 10, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_select_ok )

    self.klass.select(nowait=False)
    assert_true( self.klass._enabled )
    assert_equals( deque([None]), self.klass._select_cb )

  def test_select_when_not_enabled_with_cb(self):
    self.klass._enabled = False
    w = mock()
    expect( mock( rabbit_connection, 'Writer' ) ).returns( w )
    expect( self.klass.allow_nowait ).returns( True )
    expect( w.write_bit ).args( False )
    expect( mock(rabbit_connection, 'MethodFrame') ).args(42, 85, 10, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_select_ok )

    self.klass.select(cb='foo')
    assert_true( self.klass._enabled )
    assert_equals( deque(['foo']), self.klass._select_cb )

  def test_select_when_already_enabled(self):
    self.klass._enabled = True
    stub( self.klass.allow_nowait )
    stub( self.klass.send_frame )
    expect( self.klass.allow_nowait ).returns( True )

    assert_equals( deque(), self.klass._select_cb )
    self.klass.select()
    assert_equals( deque(), self.klass._select_cb )
  
  def test_recv_select_ok_no_cb(self):
    self.klass._select_cb = deque([None])
    self.klass._recv_select_ok('frame')
    assert_equals( deque(), self.klass._select_cb )
  
  def test_recv_select_ok_with_cb(self):
    cb = mock()
    self.klass._select_cb = deque([cb])
    expect( cb )
    self.klass._recv_select_ok('frame')
    assert_equals( deque(), self.klass._select_cb )



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai

from haigha.classes import queue_class, ProtocolClass, QueueClass
from haigha.frames import MethodFrame
from haigha.writer import Writer

from collections import deque

class QueueClassTest(Chai):

  def setUp(self):
    super(QueueClassTest,self).setUp()
    ch = mock()
    ch.channel_id = 42
    ch.logger = mock()
    self.klass = QueueClass( ch )

  def test_init(self):
    expect(ProtocolClass.__init__).args('foo', a='b')
    
    klass = QueueClass.__new__(QueueClass)
    klass.__init__('foo', a='b')

    assert_equals( 
      {
        11 : klass._recv_declare_ok,
        21 : klass._recv_bind_ok,
        31 : klass._recv_purge_ok,
        41 : klass._recv_delete_ok,
        51 : klass._recv_unbind_ok,

      }, klass.dispatch_map )
    assert_equals( deque(), self.klass._declare_cb )
    assert_equals( deque(), self.klass._bind_cb )
    assert_equals( deque(), self.klass._unbind_cb )
    assert_equals( deque(), self.klass._delete_cb )
    assert_equals( deque(), self.klass._purge_cb )

  def test_cleanup(self):
    self.klass._cleanup()
    assert_equals( None, self.klass._declare_cb )
    assert_equals( None, self.klass._bind_cb )
    assert_equals( None, self.klass._unbind_cb )
    assert_equals( None, self.klass._delete_cb )
    assert_equals( None, self.klass._purge_cb )
    assert_equals( None, self.klass._channel )
    assert_equals( None, self.klass.dispatch_map )

  def test_declare_default_args(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(queue_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( self.klass.default_ticket ).returns(w)
    expect( w.write_shortstr ).args( '' ).returns( w )
    expect( w.write_bits ).args( False, False, False, True, True ).returns( w )
    expect( w.write_table ).args( {} )
    expect( mock(queue_class, 'MethodFrame') ).args(42, 50, 10, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    stub( self.klass.channel.add_synchronous_cb )

    assert_equals( deque(), self.klass._declare_cb )
    self.klass.declare()
    assert_equals( deque(), self.klass._declare_cb )

  def test_declare_with_args_and_no_cb(self):
    w = mock()
    stub( self.klass.allow_nowait )
    expect( mock(queue_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( 'ticket' ).returns(w)
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_bits ).args( 'p', 'd', 'e', 'a', False ).returns( w )
    expect( w.write_table ).args( {'foo':'bar'} )
    expect( mock(queue_class, 'MethodFrame') ).args(42, 50, 10, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_declare_ok ).\
      returns('stuffs')

    assert_equals( deque(), self.klass._declare_cb )
    assert_equals( 'stuffs', self.klass.declare('queue', passive='p', durable='d', exclusive='e', 
      auto_delete='a', nowait=False, arguments={'foo':'bar'}, ticket='ticket') )
    assert_equals( deque([None]), self.klass._declare_cb )

  def test_declare_with_args_and_cb(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(queue_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( 'ticket' ).returns(w)
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_bits ).args( 'p', 'd', 'e', 'a', False ).returns( w )
    expect( w.write_table ).args( {'foo':'bar'} )
    expect( mock(queue_class, 'MethodFrame') ).args(42, 50, 10, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_declare_ok ).\
      returns('stuffs')

    # assert it's put in the right spot too
    self.klass._declare_cb = deque(['blargh'])
    assert_equals( 'stuffs', self.klass.declare('queue', passive='p', durable='d', exclusive='e', 
      auto_delete='a', nowait=True, arguments={'foo':'bar'}, ticket='ticket',
      cb='callback') )
    assert_equals( deque(['blargh','callback']), self.klass._declare_cb )

  def test_recv_declare_ok_with_callback(self):
    rframe = mock()
    cb = mock()
    self.klass._declare_cb.append( cb )
    self.klass._declare_cb.append( mock() ) # assert not called

    expect( rframe.args.read_shortstr ).returns( 'queue' )
    expect( rframe.args.read_long ).returns( 32 )
    expect( rframe.args.read_long ).returns( 5 )
    expect( cb ).args( 'queue', 32, 5 )

    assert_equals( (32,5), self.klass._recv_declare_ok(rframe) )
    assert_equals( 1, len(self.klass._declare_cb) )
    assert_false( cb in self.klass._declare_cb )

  def test_recv_declare_ok_without_callback(self):
    rframe = mock()
    cb = mock()
    self.klass._declare_cb.append( None )
    self.klass._declare_cb.append( cb )

    expect( rframe.args.read_shortstr ).returns( 'queue' )
    expect( rframe.args.read_long ).returns( 32 )
    expect( rframe.args.read_long ).returns( 5 )

    assert_equals( (32,5), self.klass._recv_declare_ok(rframe) )
    assert_equals( 1, len(self.klass._declare_cb) )
    assert_false( None in self.klass._declare_cb )

  def test_bind_default_args(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(queue_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( self.klass.default_ticket ).returns(w)
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_shortstr ).args( 'exchange' ).returns( w )
    expect( w.write_shortstr ).args( '' ).returns( w )
    expect( w.write_bit ).args( True ).returns( w )
    expect( w.write_table ).args( {} )
    expect( mock(queue_class, 'MethodFrame') ).args(42, 50, 20, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    stub( self.klass.channel.add_synchronous_cb )

    assert_equals( deque(), self.klass._declare_cb )
    self.klass.bind('queue', 'exchange')
    assert_equals( deque(), self.klass._declare_cb )

  def test_bind_with_args_and_no_cb(self):
    w = mock()
    stub( self.klass.allow_nowait )
    expect( mock(queue_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( 'ticket' ).returns(w)
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_shortstr ).args( 'exchange' ).returns( w )
    expect( w.write_shortstr ).args( 'route' ).returns( w )
    expect( w.write_bit ).args( False ).returns( w )
    expect( w.write_table ).args( {'foo':'bar'} )
    expect( mock(queue_class, 'MethodFrame') ).args(42, 50, 20, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_bind_ok )

    assert_equals( deque(), self.klass._bind_cb )
    self.klass.bind('queue', 'exchange', routing_key='route', nowait=False,
      arguments={'foo':'bar'}, ticket='ticket')
    assert_equals( deque([None]), self.klass._bind_cb )

  def test_bind_with_args_and_cb(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(queue_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( 'ticket' ).returns(w)
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_shortstr ).args( 'exchange' ).returns( w )
    expect( w.write_shortstr ).args( 'route' ).returns( w )
    expect( w.write_bit ).args( False ).returns( w )
    expect( w.write_table ).args( {'foo':'bar'} )
    expect( mock(queue_class, 'MethodFrame') ).args(42, 50, 20, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_bind_ok )

    self.klass._bind_cb = deque(['blargh'])
    self.klass.bind('queue', 'exchange', routing_key='route', nowait=True,
      arguments={'foo':'bar'}, ticket='ticket', cb='callback')
    assert_equals( deque(['blargh','callback']), self.klass._bind_cb )

  def test_recv_bind_ok_with_cb(self):
    cb = mock()
    self.klass._bind_cb.append( cb )
    self.klass._bind_cb.append( mock() ) # assert not called

    expect( cb )

    self.klass._recv_bind_ok( 'frame' )
    assert_equals( 1, len(self.klass._bind_cb) )
    assert_false( cb in self.klass._bind_cb )

  def test_recv_bind_ok_without_cb(self):
    self.klass._bind_cb.append( None )
    self.klass._bind_cb.append( mock() ) # assert not called

    self.klass._recv_bind_ok( 'frame' )
    assert_equals( 1, len(self.klass._bind_cb) )
    assert_false( None in self.klass._bind_cb )

  def test_unbind_default_args(self):
    w = mock()
    expect( mock(queue_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( self.klass.default_ticket ).returns(w)
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_shortstr ).args( 'exchange' ).returns( w )
    expect( w.write_shortstr ).args( '' ).returns( w )
    expect( w.write_table ).args( {} )
    expect( mock(queue_class, 'MethodFrame') ).args(42, 50, 50, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_unbind_ok )

    assert_equals( deque(), self.klass._unbind_cb )
    self.klass.unbind('queue', 'exchange')
    assert_equals( deque([None]), self.klass._unbind_cb )

  def test_unbind_with_args(self):
    w = mock()
    expect( mock(queue_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( 'ticket' ).returns(w)
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_shortstr ).args( 'exchange' ).returns( w )
    expect( w.write_shortstr ).args( 'route' ).returns( w )
    expect( w.write_table ).args( {'foo':'bar'} )
    expect( mock(queue_class, 'MethodFrame') ).args(42, 50, 50, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_unbind_ok )

    self.klass._unbind_cb = deque(['blargh'])
    self.klass.unbind('queue', 'exchange', routing_key='route', 
      arguments={'foo':'bar'}, ticket='ticket', cb='callback')
    assert_equals( deque(['blargh','callback']), self.klass._unbind_cb )

  def test_recv_unbind_ok_with_cb(self):
    cb = mock()
    self.klass._unbind_cb.append( cb )
    self.klass._unbind_cb.append( mock() ) # assert not called

    expect( cb )

    self.klass._recv_unbind_ok( 'frame' )
    assert_equals( 1, len(self.klass._unbind_cb) )
    assert_false( cb in self.klass._unbind_cb )

  def test_recv_unbind_ok_without_cb(self):
    self.klass._unbind_cb.append( None )
    self.klass._unbind_cb.append( mock() ) # assert not called

    self.klass._recv_unbind_ok( 'frame' )
    assert_equals( 1, len(self.klass._unbind_cb) )
    assert_false( None in self.klass._unbind_cb )

  def test_purge_default_args(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(queue_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( self.klass.default_ticket ).returns(w)
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_bit ).args( True )
    expect( mock(queue_class, 'MethodFrame') ).args(42, 50, 30, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    stub( self.klass.channel.add_synchronous_cb )

    assert_equals( deque(), self.klass._purge_cb )
    self.klass.purge('queue')
    assert_equals( deque(), self.klass._purge_cb )

  def test_purge_with_args_and_no_cb(self):
    w = mock()
    stub( self.klass.allow_nowait )
    expect( mock(queue_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( 'ticket' ).returns(w)
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_bit ).args( False )
    expect( mock(queue_class, 'MethodFrame') ).args(42, 50, 30, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args(
      self.klass._recv_purge_ok ).returns( 'fifty' )

    assert_equals( deque(), self.klass._purge_cb )
    assert_equals( 'fifty', self.klass.purge('queue', nowait=False, ticket='ticket') )
    assert_equals( deque([None]), self.klass._purge_cb )

  def test_purge_with_args_and_cb(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(queue_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( 'ticket' ).returns(w)
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_bit ).args( False )
    expect( mock(queue_class, 'MethodFrame') ).args(42, 50, 30, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args(
      self.klass._recv_purge_ok ).returns( 'fifty' )

    self.klass._purge_cb = deque(['blargh'])
    assert_equals( 'fifty', self.klass.purge('queue', nowait=True, ticket='ticket', cb='callback'))
    assert_equals( deque(['blargh','callback']), self.klass._purge_cb )

  def test_recv_purge_ok_with_cb(self):
    rframe = mock()
    cb = mock()
    self.klass._purge_cb.append( cb )
    self.klass._purge_cb.append( mock() ) # assert not called

    expect( rframe.args.read_long ).returns( 42 )
    expect( cb ).args( 42 )

    assert_equals( 42, self.klass._recv_purge_ok( rframe ) )
    assert_equals( 1, len(self.klass._purge_cb) )
    assert_false( cb in self.klass._purge_cb )

  def test_recv_purge_ok_without_cb(self):
    rframe = mock()
    self.klass._purge_cb.append( None )
    self.klass._purge_cb.append( mock() ) # assert not called

    expect( rframe.args.read_long ).returns( 42 )

    assert_equals( 42, self.klass._recv_purge_ok(rframe) )
    assert_equals( 1, len(self.klass._purge_cb) )
    assert_false( None in self.klass._purge_cb )

  def test_delete_default_args(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(queue_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( self.klass.default_ticket ).returns(w)
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_bits ).args( False, False, True )
    expect( mock(queue_class, 'MethodFrame') ).args(42, 50, 40, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    stub( self.klass.channel.add_synchronous_cb )

    assert_equals( deque(), self.klass._delete_cb )
    self.klass.delete('queue')
    assert_equals( deque(), self.klass._delete_cb )

  def test_delete_with_args_and_no_cb(self):
    w = mock()
    stub( self.klass.allow_nowait )
    expect( mock(queue_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( 'ticket' ).returns(w)
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_bits ).args( 'yes', 'no', False )
    expect( mock(queue_class, 'MethodFrame') ).args(42, 50, 40, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( 
      self.klass._recv_delete_ok ).returns( 'five' )

    assert_equals( deque(), self.klass._delete_cb )
    assert_equals( 'five', self.klass.delete('queue', if_unused='yes', if_empty='no', nowait=False,
      ticket='ticket') )
    assert_equals( deque([None]), self.klass._delete_cb )

  def test_delete_with_args_and_cb(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(queue_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( 'ticket' ).returns(w)
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_bits ).args( 'yes', 'no', False )
    expect( mock(queue_class, 'MethodFrame') ).args(42, 50, 40, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_delete_ok )

    self.klass._delete_cb = deque(['blargh'])
    self.klass.delete('queue', if_unused='yes', if_empty='no', nowait=True,
      ticket='ticket', cb='callback')
    assert_equals( deque(['blargh','callback']), self.klass._delete_cb )

  def test_recv_delete_ok_with_cb(self):
    rframe = mock()
    cb = mock()
    self.klass._delete_cb.append( cb )
    self.klass._delete_cb.append( mock() ) # assert not called

    expect( rframe.args.read_long ).returns( 42 )
    expect( cb ).args( 42 )

    self.klass._recv_delete_ok( rframe )
    assert_equals( 1, len(self.klass._delete_cb) )
    assert_false( cb in self.klass._delete_cb )

  def test_recv_delete_ok_without_cb(self):
    rframe = mock()
    self.klass._delete_cb.append( None )
    self.klass._delete_cb.append( mock() ) # assert not called

    expect( rframe.args.read_long ).returns( 42 )

    self.klass._recv_delete_ok( rframe )
    assert_equals( 1, len(self.klass._delete_cb) )
    assert_false( None in self.klass._delete_cb )



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

class Message(object):
  '''
  Represents an AMQP message.
  '''

  def __init__(self, body=None, delivery_info=None, **properties):
    if isinstance(body, unicode):
      if 'content_encoding' not in properties:
        properties['content_encoding'] = 'utf-8'
      body = body.encode(properties['content_encoding'])
    
    self._body = body
    self._delivery_info = delivery_info
    self._properties = properties
 
  @property
  def body(self):
    return self._body

  def __len__(self):
    return len( self._body )

  def __nonzero__(self):
    '''Have to define this because length is defined.'''
    return True

  def __eq__(self, other):
    if isinstance(other,Message):
      return self._properties == other._properties and \
             self._body == other._body
    return False

  @property
  def delivery_info(self):
    return self._delivery_info

  @property
  def properties(self):
    return self._properties

  def __str__(self):
    return "Message[body: %s, delivery_info: %s, properties: %s]"%\
      ( str(self._body).encode('string_escape'), self._delivery_info, self._properties )



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

class Transport(object):
  '''
  Base class and API for Transports
  '''

  def __init__(self, connection):
    '''
    Initialize a transport on a haigha.Connection instance.
    '''
    self._connection = connection

  @property
  def synchronous(self):
    '''Return True if this is a synchronous transport, False otherwise.'''
    # Note that subclasses must define this.
    return self._synchronous

  @property
  def connection(self):
    return self._connection

  def process_channels(self, channels):
    '''
    Process a set of channels by calling Channel.process_frames() on each. 
    Some transports may choose to do this in unique ways, such as through 
    a pool of threads.

    The default implementation will simply iterate over them and call 
    process_frames() on each.
    '''
    for channel in channels:
      channel.process_frames()

  def read(self, timeout=None):
    '''
    Read from the transport. If no data is available, should return None. The
    return value can be any data type that is supported by the haigha.Reader
    class.

    Caller passes in an optional timeout. Each transport determines how to
    implement this.
    '''
    return None

  def buffer(self, data):
    '''
    Buffer unused bytes from the input stream.
    '''

  def write(self, data):
    '''
    Write some bytes to the transport.
    '''
    
  def disconnect(self):
    '''
    Disconnect from the transport. Typically socket.close(). This call is 
    welcome to raise exceptions, which the Connection will catch.

    The transport is encouraged to allow for any pending writes to complete
    before closing the socket.
    '''



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai

from haigha.frames import content_frame, ContentFrame, Frame

class ContentFrameTest(Chai):

  def test_type(self):
    assert_equals(3, ContentFrame.type() )

  def test_payload(self):
    klass = ContentFrame(42, 'payload')
    assert_equals( 'payload', klass.payload )

  def test_parse(self):
    frame = ContentFrame.parse(42, 'payload')
    assert_true( isinstance(frame, ContentFrame) )
    assert_equals( 42, frame.channel_id )
    assert_equals( 'payload', frame.payload )

  def test_create_frames(self):
    itr = ContentFrame.create_frames(42, 'helloworld', 13)
    
    frame = itr.next()
    assert_true( isinstance(frame, ContentFrame) )
    assert_equals( 42, frame.channel_id )
    assert_equals( 'hello', frame.payload )
    
    frame = itr.next()
    assert_true( isinstance(frame, ContentFrame) )
    assert_equals( 42, frame.channel_id )
    assert_equals( 'world', frame.payload )

    assert_raises(StopIteration, itr.next)

  def test_init(self):
    expect(Frame.__init__).args( is_a(ContentFrame), 42 )
    frame = ContentFrame(42, 'payload')
    assert_equals( 'payload', frame._payload )

  def test_str(self):
    # Test both branches but don't assert the actual content because its
    # not worth it
    frame = ContentFrame(42, 'payload')
    str(frame)

    frame = ContentFrame(42, 8675309)
    str(frame)

  def test_write_frame(self):
    w = mock()
    expect( mock(content_frame, 'Writer') ).args('buffer').returns( w )
    expect( w.write_octet ).args( 3 ).returns( w )
    expect( w.write_short ).args( 42 ).returns( w )
    expect( w.write_long ).args( 5 ).returns( w )
    expect( w.write ).args( 'hello' ).returns( w )
    expect( w.write_octet ).args( 0xce )

    frame = ContentFrame(42, 'hello')
    frame.write_frame( 'buffer' )



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai
import struct
from collections import deque

from haigha.frames import frame
from haigha.frames import Frame
from haigha.reader import Reader

class FrameTest(Chai):

  def test_register(self):
    class DummyFrame(Frame):
      @classmethod
      def type(self): return 42

    assertEquals( None, Frame._frame_type_map.get(42) )
    DummyFrame.register()
    assertEquals( DummyFrame, Frame._frame_type_map[42] )

  def test_type_raises_not_implemented(self):
    assertRaises( NotImplementedError, Frame.type )

  def test_read_frames_reads_until_buffer_underflow(self):
    reader = mock()

    expect( reader.tell ).returns( 0 )
    expect( Frame._read_frame ).args(reader).returns('frame1')

    expect( reader.tell ).returns( 2 )
    expect( Frame._read_frame ).args(reader).returns('frame2')

    expect( reader.tell ).returns( 3 )
    expect( Frame._read_frame ).args(reader).raises( Reader.BufferUnderflow )

    expect( reader.seek ).args( 3 )

    assertEquals( deque(['frame1','frame2']), Frame.read_frames(reader) )

  def test_read_frames_handles_reader_errors(self):
    reader = mock()
    self.mock( Frame, '_read_frame' )
    
    expect( reader.tell ).returns( 0 )
    expect( Frame._read_frame ).args(reader).raises( Reader.ReaderError("bad!") )

    assertRaises( Frame.FormatError, Frame.read_frames, reader )

  def test_read_frames_handles_struct_errors(self):
    reader = mock()
    self.mock( Frame, '_read_frame' )
    
    expect( reader.tell ).returns( 0 )
    expect( Frame._read_frame ).args(reader).raises( struct.error("bad!") )

    self.assertRaises( Frame.FormatError, Frame.read_frames, reader )

  def test_read_frame_on_full_frame(self):
    class FrameReader(Frame):
      @classmethod
      def type(self): return 45

      @classmethod
      def parse(self, channel_id, payload):
        return 'no_frame'
    FrameReader.register()

    self.mock( frame, 'Reader' )
    reader = self.mock()
    payload = self.mock()

    expect( reader.read_octet ).returns( 45 ) # frame type
    expect( reader.read_short ).returns( 32 ) # channel id
    expect( reader.read_long ).returns( 42 )  # size
    
    expect( reader.tell ).returns( 5 )
    expect( frame.Reader ).args(reader, 5, 42).returns( payload )
    expect( reader.seek ).args( 42, 1 )

    expect( reader.read_octet ).returns( 0xce )
    expect( FrameReader.parse ).args( 32, payload ).returns( 'a_frame' )

    assertEquals( 'a_frame', Frame._read_frame(reader) )

  def test_read_frame_raises_bufferunderflow_when_incomplete_payload(self):
    self.mock( frame, 'Reader' )
    reader = self.mock()

    expect( reader.read_octet ).returns( 45 ) # frame type
    expect( reader.read_short ).returns( 32 ) # channel id
    expect( reader.read_long ).returns( 42 )  # size
    
    expect( reader.tell ).returns( 5 )
    expect( frame.Reader ).args(reader, 5, 42).returns( 'payload' )
    expect( reader.seek ).args( 42, 1 )

    expect( reader.read_octet ).raises( Reader.BufferUnderflow )
    assert_raises( Reader.BufferUnderflow, Frame._read_frame, reader )

  def test_read_frame_raises_formaterror_if_bad_footer(self):
    self.mock( frame, 'Reader' )
    reader = self.mock()

    expect( reader.read_octet ).returns( 45 ) # frame type
    expect( reader.read_short ).returns( 32 ) # channel id
    expect( reader.read_long ).returns( 42 )  # size
    
    expect( reader.tell ).returns( 5 )
    expect( frame.Reader ).args(reader, 5, 42).returns( 'payload' )
    expect( reader.seek ).args( 42, 1 )
    expect( reader.read_octet ).returns( 0xff )

    assert_raises( Frame.FormatError, Frame._read_frame, reader )

  def test_read_frame_raises_invalidframetype_for_unregistered_frame_type(self):
    self.mock( frame, 'Reader' )
    reader = self.mock()
    payload = self.mock()

    expect( reader.read_octet ).returns( 54 ) # frame type
    expect( reader.read_short ).returns( 32 ) # channel id
    expect( reader.read_long ).returns( 42 )  # size
    
    expect( reader.tell ).returns( 5 )
    expect( frame.Reader ).args(reader, 5, 42).returns( payload )
    expect( reader.seek ).args( 42, 1 )

    expect( reader.read_octet ).returns( 0xce )

    assertRaises( Frame.InvalidFrameType, Frame._read_frame, reader )
  
  def test_parse_raises_not_implemented(self):
    assertRaises( NotImplementedError, Frame.parse, 'channel_id', 'payload' )

  def test_properties(self):
    frame = Frame('channel_id')
    assert_equals('channel_id', frame.channel_id)

  def test_str(self):
    frame = Frame(42)
    assert_equals('Frame[channel: 42]', str(frame))

  def test_repr(self):
    expect(Frame.__str__).returns('foo')
    frame = Frame(42)
    assert_equals('foo', repr(frame))

  def test_write_frame(self):
    frame = Frame(42)
    assert_raises( NotImplementedError, frame.write_frame, 'stream' )



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.transports import Transport

try:
  from eventsocket import EventSocket
  import event
except ImportError:
  print 'Failed to load EventSocket and event modules'
  EventSocket = None
  event = None

class EventTransport(Transport):
  '''
  Transport using libevent-based EventSocket.
  '''
  
  def __init__(self, *args):
    super(EventTransport,self).__init__(*args)
    self._synchronous = False

  ###
  ### EventSocket callbacks
  ###
  def _sock_close_cb(self, sock):
    self._connection.transport_closed(
      msg='socket to %s closed unexpectedly'%(self._host),
    )

  def _sock_error_cb(self, sock, msg, exception=None):
    self._connection.transport_closed(
      msg='error on connection to %s: %s'%(self._host, msg)
    )

  def _sock_read_cb(self, sock):
    self.connection.read_frames()
  
  ###
  ### Transport API
  ###
  def connect(self, (host,port)):
    '''
    Connect assuming a host and port tuple. Implemented as non-blocking, and
    will close the transport if there's an error
    '''
    self._host = "%s:%s"%(host,port)
    self._sock = EventSocket(
      read_cb=self._sock_read_cb,
      close_cb=self._sock_close_cb, 
      error_cb=self._sock_error_cb,
      debug=self.connection.debug, 
      logger=self.connection.logger )
    if self.connection._sock_opts:
      for k,v in self.connection._sock_opts.iteritems():
        family,type = k
        self._sock.setsockopt(family, type, v)
    self._sock.setblocking( False )
    self._sock.connect( (host,port), timeout=self.connection._connect_timeout )
    self._heartbeat_timeout = None

  def read(self, timeout=None):
    '''
    Read from the transport. If no data is available, should return None. The
    timeout is ignored as this returns only data that has already been buffered
    locally.
    '''
    # NOTE: copying over this comment from Connection, because there is
    # knowledge captured here, even if the details are stale
    # Because of the timer callback to dataRead when we re-buffered, there's a
    # chance that in between we've lost the socket.  If that's the case, just
    # silently return as some code elsewhere would have already notified us.
    # That bug could be fixed by improving the message reading so that we consume
    # all possible messages and ensure that only a partial message was rebuffered,
    # so that we can rely on the next read event to read the subsequent message.
    if not hasattr(self,'_sock'):
      return None

    # This is sort of a hack because we're faking that data is ready, but it
    # works for purposes of supporting timeouts
    if timeout:
      if self._heartbeat_timeout:
        self._heartbeat_timeout.delete()
      self._heartbeat_timeout = \
        event.timeout( timeout, self._sock_read_cb, self._sock )
    elif self._heartbeat_timeout:
      self._heartbeat_timeout.delete()
      self._heartbeat_timeout = None
      
    return self._sock.read()

  def buffer(self, data):
    '''
    Buffer unused bytes from the input stream.
    '''
    if not hasattr(self,'_sock'):
      return None
    self._sock.buffer( data )

  def write(self, data):
    '''
    Write some bytes to the transport.
    '''
    if not hasattr(self,'_sock'):
      return
    self._sock.write( data )
    
  def disconnect(self):
    '''
    Disconnect from the transport. Typically socket.close(). This call is 
    welcome to raise exceptions, which the Connection will catch.

    The transport is encouraged to allow for any pending writes to complete
    before closing the socket.
    '''
    if not hasattr(self,'_sock'):
      return

    # TODO: If there are bytes left on the output, queue the close for later.
    self._sock.close_cb = None
    self._sock.close()



'''
Copyright (c) 2011, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.writer import Writer
from haigha.frames.frame import Frame

class ContentFrame(Frame):
  '''
  Frame for reading in content.
  '''

  @classmethod
  def type(cls):
    return 3

  @property
  def payload(self):
    return self._payload

  @classmethod
  def parse(self, channel_id, payload):
    return ContentFrame( channel_id, payload )

  @classmethod
  def create_frames(self, channel_id, buf, frame_max):
    '''
    A generator which will create frames from a buffer given a max
    frame size.
    '''
    size = frame_max - 8   # 8 bytes overhead for frame header and footer
    offset = 0
    while True:
      payload = buf[offset:(offset+size)]
      if len(payload)==0: break
      offset += size
      
      yield ContentFrame(channel_id, payload)
      if offset >= len(buf): break
    
  def __init__(self, channel_id, payload):
    Frame.__init__(self, channel_id)
    self._payload = payload
  
  def __str__(self):
    if isinstance(self._payload, str):
      payload = ''.join( ['\\x%s'%(c.encode('hex')) for c in self._payload] )
    else:
      payload = str(self._payload)
    
    return "%s[channel: %d, payload: %s]"%( self.__class__.__name__, self.channel_id, payload )

  def write_frame(self, buf):
    '''
    Write the frame into an existing buffer.
    '''
    writer = Writer( buf )

    writer.write_octet( self.type() ).\
      write_short(self.channel_id).\
      write_long( len(self._payload) ).\
      write( self._payload ).\
      write_octet( 0xce )


ContentFrame.register()

